""" Integration Test for Data Rest API
"""
from copy import copy

from mock import patch
from rest_framework import status

from core_main_app.components.data.models import Data
from core_main_app.components.workspace.models import Workspace
from core_main_app.rest.data import views as data_rest_views
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.integration_tests.integration_base_transaction_test_case import (
    MongoIntegrationTransactionTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from tests.components.data.fixtures.fixtures import (
    DataFixtures,
    QueryDataFixtures,
    AccessControlDataFixture,
)
from tests.components.user.fixtures.fixtures import UserFixtures
from core_main_app.components.workspace import api as workspace_api
from core_main_app.components.data import api as data_api

fixture_data = DataFixtures()
fixture_data_query = QueryDataFixtures()
fixture_data_workspace = AccessControlDataFixture()


class TestDataListbyWorkspace(MongoIntegrationBaseTestCase):
    fixture = fixture_data_workspace

    def setUp(self):
        super(TestDataListbyWorkspace, self).setUp()

    def test_get_filtered_by_correct_workspace_returns_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(),
            user,
            data={"workspace": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(len(response.data), 2)

    def test_get_filtered_by_incorrect_workspace_returns_no_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(),
            user,
            data={"workspace": "5da8ac0622301d1240f2551b"},
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_get_all_by_correct_workspace_returns_data(self):
        # Arrange
        user = create_mock_user("1", is_superuser=True)

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataListByWorkspace.as_view(),
            user,
            param={"workspace_id": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(len(response.data), 2)


class TestDataList(MongoIntegrationBaseTestCase):
    fixture = fixture_data

    def setUp(self):
        super(TestDataList, self).setUp()

    @patch.object(Data, "xml_content")
    def test_get_returns_http_200(self, mock_xml_content):
        # Arrange
        user = create_mock_user("1")
        mock_xml_content.return_value = "content"

        # Act
        response = RequestMock.do_request_get(data_rest_views.DataList.as_view(), user)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_all_user_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(data_rest_views.DataList.as_view(), user)

        # Assert
        self.assertEqual(len(response.data), 2)

    def test_get_filtered_by_correct_title_returns_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(),
            user,
            data={"title": self.fixture.data_1.title},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_title_returns_no_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(), user, data={"title": "bad title"}
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_get_filtered_by_correct_template_returns_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(),
            user,
            data={"template": self.fixture.data_1.template},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_template_returns_no_data(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataList.as_view(),
            user,
            data={"template": "507f1f77bcf86cd799439011"},
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_post_data_missing_field_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")
        mock_data = {
            "template": str(self.fixture.template.id),
            "user_id": "1",
            "xml_content": "<tag></tag>",
        }

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.DataList.as_view(), user, data=mock_data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_post_data_incorrect_template_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")
        mock_data = {
            "template": "507f1f77bcf86cd799439011",
            "user_id": "1",
            "title": "new data",
            "xml_content": "<tag></tag>",
        }

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.DataList.as_view(), user, data=mock_data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestDataDetail(MongoIntegrationBaseTestCase):
    fixture = fixture_data

    def setUp(self):
        super(TestDataDetail, self).setUp()

    @patch.object(Data, "xml_content")
    def test_get_returns_http_200(self, mock_xml_content):
        # Arrange
        user = create_mock_user("1")
        mock_xml_content.return_value = "content"

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": self.fixture.data_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Data, "xml_content")
    def test_get_returns_data(self, mock_xml_content):
        # Arrange
        user = create_mock_user("1")
        mock_xml_content.return_value = "content"

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": self.fixture.data_1.id},
        )

        # Assert
        self.assertEqual(response.data["title"], self.fixture.data_1.title)

    @patch.object(Data, "xml_content")
    def test_get_data_containing_ascii_returns_data(self, mock_xml_content):
        # Arrange
        user = create_mock_user("1")
        mock_xml_content.return_value = "\xc3te\xc3"

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": self.fixture.data_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": "507f1f77bcf86cd799439011"},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_delete(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": "507f1f77bcf86cd799439011"},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataDetail.as_view(),
            user,
            param={"pk": "507f1f77bcf86cd799439011"},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_wrong_template_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataDetail.as_view(),
            user,
            data={"template": "507f1f77bcf86cd799439011"},
            param={"pk": self.fixture.data_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestDataDownload(MongoIntegrationBaseTestCase):
    fixture = fixture_data

    def setUp(self):
        super(TestDataDownload, self).setUp()

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDownload.as_view(),
            user,
            param={"pk": self.fixture.data_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataDownload.as_view(),
            user,
            param={"pk": "507f1f77bcf86cd799439011"},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TestExecuteLocalQueryView(MongoIntegrationBaseTestCase):
    fixture = fixture_data_query

    def setUp(self):
        super(TestExecuteLocalQueryView, self).setUp()
        # FIXME: unable to test paginated results (mocked queryset.count always returns 0)
        self.data = {"all": "true"}
        # create user with superuser access to skip access control
        self.user = create_mock_user("1", is_superuser=True)

    def test_post_query_string_one_data_returns_http_200(self):
        # Arrange
        self.data.update({"query": '{"root.element": "value"}'})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_query_string_one_data_returns_one_data(self):
        # Arrange
        self.data.update({"query": '{"root.element": "value"}'})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_post_query_string_two_data_returns_two_data(self):
        # Arrange
        self.data.update(
            {"query": '{"$or": [{"root.element": "value"}, {"root.element":"value2"}]}'}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)

    def test_post_empty_query_string_filter_by_templates_returns_all_data_of_the_template(
        self,
    ):
        # Arrange
        self.data.update(
            {"query": "{}", "templates": [{"id": str(self.fixture.template.id)}]}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)

    def test_post_query_string_filtered_by_templates_returns_one_data(self):
        # Arrange
        self.data.update(
            {
                "query": '{"root.element": "value"}',
                "templates": [{"id": str(self.fixture.template.id)}],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_post_query_string_filtered_by_wrong_template_id_returns_no_data(self):
        # Arrange
        self.data.update(
            {
                "query": '{"root.element": "value"}',
                "templates": [{"id": "507f1f77bcf86cd799439011"}],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_post_query_json_returns_http_200(self):
        # Arrange
        self.data.update({"query": {}})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_query_json_int_returns_data_1(self):
        # Arrange
        self.data.update({"query": {"root.complex.child2": 0}})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_post_query_json_string_returns_data_1(self):
        # Arrange
        self.data.update({"query": {"root.complex.child1": "test"}})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_post_query_json_regex_returns_all_data(self):
        # Arrange
        self.data.update({"query": {"root.element": "/.*/"}})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), len(self.fixture.data_collection))

    def test_post_query_json_or_operator_returns_data_1_and_data_2(self):
        # Arrange
        self.data.update(
            {"query": {"$or": [{"root.element": "value"}, {"root.element": "value2"}]}}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), len(self.fixture.data_collection))

    def test_post_query_json_and_operator_returns_data_1(self):
        # Arrange
        self.data.update(
            {
                "query": {
                    "$and": [
                        {"root.complex.child2": {"$lt": 1}},
                        {"root.complex.child2": {"$gte": 0}},
                    ]
                }
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_post_query_json_element_match_operator_returns_data_1(self):
        # Arrange
        self.data.update(
            {"query": {"root.list": {"$elemMatch": {"element_list_1": 1}}}}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)


class TestExecuteLocalQueryViewWorkspaceCase(MongoIntegrationTransactionTestCase):
    fixture = fixture_data_workspace

    def setUp(self):
        super(TestExecuteLocalQueryViewWorkspaceCase, self).setUp()

        self.data = {"all": "true"}

        # create user with superuser access to skip access control
        self.user = create_mock_user("1", is_superuser=True)

        self.user2 = UserFixtures().create_user()

        self.user2.id = self.fixture.data_4.user_id

    def test_post_empty_query_with_no_specific_workspace_returns_all_accessible_data_as_superuser(
        self,
    ):

        # Arrange
        self.data.update({"query": "{}", "workspaces": []})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 5)

    def test_post_empty_query_with_no_specific_workspace_returns_all_accessible_data_as_user(
        self,
    ):

        # Arrange
        self.data.update({"query": "{}", "workspaces": []})
        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user2, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)

        for data in response.data:
            self.assertEqual(data["user_id"], str(self.user2.id))

    def test_post_empty_query_string_filter_by_one_workspace_returns_all_data_of_the_workspace(
        self,
    ):

        # Arrange
        self.data.update(
            {"query": "{}", "workspaces": [{"id": str(self.fixture.workspace_1.id)}]}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)

        for data in response.data:
            self.assertEqual(data["workspace"], str(self.fixture.workspace_1.id))

    def test_post_empty_query_string_filter_by_workspaces_returns_all_data_of_those_workspaces(
        self,
    ):

        # Arrange
        self.data.update(
            {
                "query": "{}",
                "workspaces": [
                    {"id": str(self.fixture.workspace_1.id)},
                    {"id": str(self.fixture.workspace_2.id)},
                ],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 3)

        list_ids = [str(self.fixture.workspace_1.id), str(self.fixture.workspace_2.id)]
        for data in response.data:
            self.assertIn(data["workspace"], list_ids)

    def test_post_query_filter_by_correct_and_wrong_workspaces_returns_data_from_correct_workspace_only_as_superuser(
        self,
    ):

        # Arrange
        self.data.update(
            {
                "query": "{}",
                "workspaces": [
                    {"id": "507f1f77bcf86cd799439011"},
                    {"id": str(self.fixture.workspace_2.id)},
                ],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

        for data in response.data:
            self.assertEqual(data["workspace"], str(self.fixture.workspace_2.id))

    def test_post_query_filter_by_correct_and_wrong_workspaces_returns_data_from_correct_workspace_only_as_user(
        self,
    ):

        # Arrange
        self.data.update(
            {
                "query": "{}",
                "workspaces": [
                    {"id": str(self.fixture.workspace_1.id)},
                    {"id": str(self.fixture.workspace_2.id)},
                ],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user2, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

        list_ids = [str(self.fixture.workspace_2.id)]
        for data in response.data:
            self.assertIn(data["workspace"], list_ids)

    def test_post_query_string_filter_by_workspace_returns_data_3(self):

        # Arrange
        self.data.update(
            {
                "query": '{"root.element": "value2"}',
                "workspaces": [{"id": str(self.fixture.workspace_1.id)}],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)

        for data in response.data:
            self.assertEqual(data["id"], str(self.fixture.data_3.id))
            self.assertEqual(data["workspace"], str(self.fixture.workspace_1.id))

    def test_post_query_string_filter_by_workspace_returns_no_data(self):

        # Arrange
        self.data.update(
            {
                "query": '{"root.element": "value2"}',
                "workspaces": [{"id": str(self.fixture.workspace_2.id)}],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_post_query_string_filter_by_private_workspace_returns_all_data_with_no_workspace_as_superuser(
        self,
    ):

        # Arrange
        self.data.update({"query": "{}", "workspaces": [{"id": "None"}]})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)
        for data in response.data:
            self.assertEqual(data["workspace"], None)

    def test_post_query_string_filter_by_private_workspace_returns_all_data_with_no_workspace_as_user(
        self,
    ):

        # Arrange
        self.data.update({"query": "{}", "workspaces": [{"id": "None"}]})

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user2, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 1)
        for data in response.data:
            self.assertEqual(data["workspace"], None)
            self.assertEqual(data["user_id"], self.user2.id)

    def test_post_query_filter_by_private_and_normal_workspaces_returns_all_data_of_those_workspaces_as_superuser(
        self,
    ):

        # Arrange
        self.data.update(
            {
                "query": "{}",
                "workspaces": [
                    {"id": "None"},
                    {"id": str(self.fixture.workspace_2.id)},
                ],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 3)

        list_ids = [None, str(self.fixture.workspace_2.id)]
        for data in response.data:
            self.assertIn(data["workspace"], list_ids)

    def test_post_query_filter_by_private_and_normal_workspaces_returns_all_data_of_those_workspaces_as_user(
        self,
    ):

        # Arrange
        self.data.update(
            {
                "query": "{}",
                "workspaces": [
                    {"id": "None"},
                    {"id": str(self.fixture.workspace_2.id)},
                ],
            }
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user2, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 2)

        list_ids = [None, str(self.fixture.workspace_2.id)]
        for data in response.data:
            self.assertIn(data["workspace"], list_ids)

    def test_post_filtered_by_wrong_workspace_id_returns_no_data(self):
        # Arrange
        self.data.update(
            {"query": "{}", "workspaces": [{"id": "507f1f77bcf86cd799439011"}]}
        )

        # Act
        response = RequestMock.do_request_post(
            data_rest_views.ExecuteLocalQueryView.as_view(), self.user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data), 0)


class TestDataAssign(MongoIntegrationBaseTestCase):
    fixture = fixture_data_workspace

    @patch.object(Workspace, "get_by_id")
    @patch.object(Data, "get_by_id")
    def test_get_returns_http_200(self, data_get_by_id, workspace_get_by_id):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user = create_mock_user(data.user_id, is_superuser=True)
        data_get_by_id.return_value = data
        workspace_get_by_id.return_value = self.fixture.workspace_1

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataAssign.as_view(),
            user,
            param={"pk": data.id, "workspace_id": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Workspace, "get_by_id")
    @patch.object(Data, "get_by_id")
    def test_assign_data_to_workspace_updates_workspace(
        self, data_get_by_id, workspace_get_by_id
    ):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user = create_mock_user(data.user_id, is_superuser=True)
        data_get_by_id.return_value = data
        workspace_get_by_id.return_value = self.fixture.workspace_2

        # Act
        RequestMock.do_request_patch(
            data_rest_views.DataAssign.as_view(),
            user,
            param={"pk": data.id, "workspace_id": self.fixture.workspace_2.id},
        )

        # Assert
        self.assertEqual(str(data.workspace.id), str(self.fixture.workspace_2.id))

    @patch.object(Workspace, "get_by_id")
    @patch.object(Data, "get_by_id")
    def test_assign_data_to_workspace_returns_http_200(
        self, data_get_by_id, workspace_get_by_id
    ):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user = create_mock_user(data.user_id, is_superuser=True)
        data_get_by_id.return_value = data
        workspace_get_by_id.return_value = self.fixture.workspace_2

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataAssign.as_view(),
            user,
            param={"pk": data.id, "workspace_id": self.fixture.workspace_2.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Workspace, "get_by_id")
    def test_assign_bad_data_id_returns_http_404(self, workspace_get_by_id):
        # Arrange
        fake_data_id = "507f1f77bcf86cd799439011"
        user = create_mock_user("1", is_superuser=True)
        workspace_get_by_id.return_value = self.fixture.workspace_2

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataAssign.as_view(),
            user,
            param={"pk": fake_data_id, "workspace_id": self.fixture.workspace_2.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch.object(Data, "get_by_id")
    def test_assign_bad_workspace_id_returns_http_404(self, data_get_by_id):
        # Arrange
        fake_workspace_id = "507f1f77bcf86cd799439011"
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user = create_mock_user(data.user_id, is_superuser=True)
        data_get_by_id.return_value = data

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataAssign.as_view(),
            user,
            param={"pk": data.id, "workspace_id": fake_workspace_id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TestDataChangeOwner(MongoIntegrationBaseTestCase):
    fixture = fixture_data_workspace

    @patch("core_main_app.components.user.api.get_user_by_id")
    @patch.object(Data, "get_by_id")
    def test_get_returns_http_200_if_user_is_superuser(
        self, data_get_by_id, user_get_by_id
    ):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user_request = create_mock_user("65467", is_staff=True, is_superuser=True)
        user_new_owner = create_mock_user("123")
        data_get_by_id.return_value = data
        user_get_by_id.return_value = user_new_owner

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataChangeOwner.as_view(),
            user_request,
            param={"pk": data.id, "user_id": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch("core_main_app.components.user.api.get_user_by_id")
    @patch.object(Data, "get_by_id")
    def test_get_returns_http_200_if_user_is_not_superuser_but_have_access_to_the_data(
        self, data_get_by_id, user_get_by_id
    ):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user_request = create_mock_user("1", is_staff=True)
        user_new_owner = create_mock_user("123")
        data_get_by_id.return_value = data
        user_get_by_id.return_value = user_new_owner

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataChangeOwner.as_view(),
            user_request,
            param={"pk": data.id, "user_id": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch(
        "core_main_app.components.workspace.api.get_all_workspaces_with_read_access_by_user"
    )
    @patch("core_main_app.components.user.api.get_user_by_id")
    @patch.object(Data, "get_by_id")
    def test_get_returns_http_403_if_user_is_not_superuser_or_have_no_access_to_the_data(
        self, data_get_by_id, user_get_by_id, read_access_mock
    ):
        # Arrange
        data = self.fixture.data_collection[self.fixture.USER_1_WORKSPACE_1]
        user_request = create_mock_user("65467", is_staff=True)
        user_new_owner = create_mock_user("123")
        data_get_by_id.return_value = data
        user_get_by_id.return_value = user_new_owner
        read_access_mock.return_value = []

        # Act
        response = RequestMock.do_request_patch(
            data_rest_views.DataChangeOwner.as_view(),
            user_request,
            param={"pk": data.id, "user_id": self.fixture.workspace_1.id},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class TestDataPermissions(MongoIntegrationTransactionTestCase):
    fixture = fixture_data_workspace
    admin_user = UserFixtures().create_super_user("admin")

    def setUp(self):
        """Insert needed data.

        Returns:

        """
        super().setUp()
        self.fixture.insert_data()
        self.fixture.generate_workspace_with_perm()

    def test_get_returns_correct_permissions_if_user_is_superuser(self):
        # Arrange
        user_request = create_mock_user("1", is_superuser=True)
        data = self.fixture.data_4

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = True
        self.assertEqual(response.data, excepted_result)

    def test_get_returns_correct_permissions_if_user_is_anonymous(self):
        # Arrange
        user_request = create_mock_user("1", is_anonymous=True)
        data = self.fixture.data_4

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = False
        self.assertEqual(response.data, excepted_result)

    @patch.object(data_api, "get_by_id")
    def test_get_returns_correct_permissions_if_user_is_not_owner(self, data_get_by_id):
        # Arrange
        user_request = UserFixtures().create_user("no_owner_user", is_staff=True)
        data = self.fixture.data_4
        data_get_by_id.return_value = data

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = False
        self.assertEqual(response.data, excepted_result)

    @patch.object(data_api, "get_by_id")
    def test_get_returns_correct_permissions_if_user_is_owner(self, data_get_by_id):
        # Arrange
        user_request = UserFixtures().create_user("owner_user", is_staff=True)
        data = copy(self.fixture.data_3)  # do not alter the fixture object
        data.user_id = str(user_request.id)
        data.workspace = None
        data_get_by_id.return_value = data

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = True
        self.assertEqual(response.data, excepted_result)

    @patch.object(data_api, "get_by_id")
    def test_get_if_user_has_read_permission_in_workspace(self, data_get_by_id):
        # Arrange
        user_request = UserFixtures().create_user("read_user")
        data = self.fixture.data_collection[self.fixture.USER_2_WORKSPACE_2]
        data_get_by_id.return_value = data
        workspace_api.add_user_read_access_to_workspace(
            self.fixture.workspace_2, user_request, self.admin_user
        )

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = False
        self.assertEqual(response.data, excepted_result)

    @patch.object(data_api, "get_by_id")
    def test_get_if_user_has_write_permission_in_workspace(self, data_get_by_id):
        # Arrange
        user_request = UserFixtures().create_user("write_user")
        data = self.fixture.data_collection[self.fixture.USER_2_WORKSPACE_2]
        data_get_by_id.return_value = data
        workspace_api.add_user_write_access_to_workspace(
            self.fixture.workspace_2, user_request, self.admin_user
        )

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = True
        self.assertEqual(response.data, excepted_result)

    @patch.object(data_api, "get_by_id")
    def test_get_if_user_has_write_read_permission_in_workspace(self, data_get_by_id):
        # Arrange
        user_request = UserFixtures().create_user("rw_user")
        data = self.fixture.data_collection[self.fixture.USER_2_WORKSPACE_2]
        data_get_by_id.return_value = data
        workspace_api.add_user_read_access_to_workspace(
            self.fixture.workspace_2, user_request, self.admin_user
        )
        workspace_api.add_user_write_access_to_workspace(
            self.fixture.workspace_2, user_request, self.admin_user
        )

        # Act
        response = RequestMock.do_request_get(
            data_rest_views.DataPermissions.as_view(),
            user_request,
            data={"ids": f'["{data.id}"]'},
        )

        # Assert
        excepted_result = {}
        excepted_result[str(data.id)] = True
        self.assertEqual(response.data, excepted_result)



""" Access control testing
"""

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.template import api as template_api
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import create_mock_request
from tests.components.template.fixtures.fixtures import AccessControlTemplateFixture

fixture_template = AccessControlTemplateFixture()


# FIXME: missing tests where CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT is True


class TestTemplateUpsert(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_upsert_user_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.upsert(self.fixture.user1_template, request=mock_request)

    def test_upsert_global_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.upsert(self.fixture.global_template, request=mock_request)

    def test_upsert_own_template_as_user_saves(self):
        mock_request = create_mock_request(user=self.user1)
        template_api.upsert(self.fixture.user1_template, request=mock_request)

    def test_upsert_other_users_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.upsert(self.fixture.user2_template, request=mock_request)

    def test_upsert_global_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.upsert(self.fixture.global_template, request=mock_request)

    def test_upsert_own_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.upsert(self.fixture.user1_template, request=mock_request)

    def test_upsert_other_users_template_as_staff_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            template_api.upsert(self.fixture.user2_template, request=mock_request)

    def test_upsert_global_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.upsert(self.fixture.global_template, request=mock_request)

    def test_upsert_own_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.upsert(self.fixture.user1_template, request=mock_request)

    def test_upsert_other_users_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.upsert(self.fixture.user2_template, request=mock_request)

    def test_upsert_global_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.upsert(self.fixture.global_template, request=mock_request)


class TestTemplateSetDisplayName(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_set_display_name_user_template_as_anonymous_raises_access_control_error(
        self,
    ):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.set_display_name(
                self.fixture.user1_template, "new_name", request=mock_request
            )

    def test_set_display_name_global_template_as_anonymous_raises_access_control_error(
        self,
    ):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.set_display_name(
                self.fixture.global_template, "new_name", request=mock_request
            )

    def test_set_display_name_own_template_as_user_saves(self):
        mock_request = create_mock_request(user=self.user1)
        template_api.set_display_name(
            self.fixture.user1_template, "new_name", request=mock_request
        )

    def test_set_display_name_other_users_template_as_user_raises_access_control_error(
        self,
    ):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.set_display_name(
                self.fixture.user2_template, "new_name", request=mock_request
            )

    def test_set_display_name_global_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.set_display_name(
                self.fixture.global_template, "new_name", request=mock_request
            )

    def test_set_display_name_own_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.set_display_name(
            self.fixture.user1_template, "new_name", request=mock_request
        )

    def test_set_display_name_other_users_template_as_staff_raises_access_control_error(
        self,
    ):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            template_api.set_display_name(
                self.fixture.user2_template, "new_name", request=mock_request
            )

    def test_set_display_name_global_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.set_display_name(
            self.fixture.global_template, "new_name", request=mock_request
        )

    def test_set_display_name_own_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.set_display_name(
            self.fixture.user1_template, "new_name", request=mock_request
        )

    def test_set_display_name_other_users_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.set_display_name(
            self.fixture.user2_template, "new_name", request=mock_request
        )

    def test_set_display_name_global_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.set_display_name(
            self.fixture.global_template, "new_name", request=mock_request
        )


class TestTemplateGet(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_get_user_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.get(self.fixture.user1_template.id, request=mock_request)

    def test_get_global_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.get(self.fixture.global_template.id, request=mock_request)

    def test_get_own_template_as_user_returns_template(self):
        mock_request = create_mock_request(user=self.user1)
        template = template_api.get(
            self.fixture.user1_template.id, request=mock_request
        )
        self.assertEqual(template, self.fixture.user1_template)

    def test_global_template_as_user_returns_template(self):
        mock_request = create_mock_request(user=self.user1)
        template = template_api.get(
            self.fixture.global_template.id, request=mock_request
        )
        self.assertEqual(template, self.fixture.global_template)

    def test_get_other_users_template_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.get(self.fixture.user2_template.id, request=mock_request)

    def test_get_any_template_as_superuser_returns_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        template = template_api.get(
            self.fixture.user1_template.id, request=mock_request
        )
        self.assertEqual(template, self.fixture.user1_template)
        template = template_api.get(
            self.fixture.user2_template.id, request=mock_request
        )
        self.assertEqual(template, self.fixture.user2_template)
        template = template_api.get(
            self.fixture.global_template.id, request=mock_request
        )
        self.assertEqual(template, self.fixture.global_template)

    def test_get_other_users_template_as_staff_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            template_api.get(self.fixture.user2_template.id, request=mock_request)


class TestTemplateGetAllAccessibleByIdList(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()
        self.template_id_list = [
            str(self.fixture.user1_template.id),
            str(self.fixture.user2_template.id),
            str(self.fixture.global_template.id),
        ]

    def test_get_all_accessible_by_id_list_as_anonymous_returns_nothing(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        templates = template_api.get_all_accessible_by_id_list(
            self.template_id_list, request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_id_list_as_user_returns_accessible_templates(self):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_id_list(
            self.template_id_list, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_id_list_as_staff_returns_accessible_templates(self):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_id_list(
            self.template_id_list, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_id_list_as_superuser_returns_accessible_templates(
        self,
    ):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_id_list(
            self.template_id_list, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))


class TestTemplateGetAllByHash(MongoIntegrationBaseTestCase):
    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_get_all_accessible_by_hash_as_anonymous_does_not_return_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.anonymous_user)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user1_template.hash, request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_as_anonymous_does_not_return_global(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.global_template.hash, request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_as_user_returns_user_template(self):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user1_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_as_user_returns_global_template(self):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.global_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_as_user_does_not_return_other_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user2_template.hash, request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_as_staff_returns_user_template(self):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user1_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_as_staff_returns_global_template(self):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.global_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_as_staff_does_not_return_other_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user2_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_as_superuser_returns_user_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user1_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_as_superuser_returns_global_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.global_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_as_superuser_returns_other_user_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash(
            self.fixture.user2_template.hash, request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))


class TestTemplateGetAllByHashList(MongoIntegrationBaseTestCase):
    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_get_all_accessible_by_hash_list_as_anonymous_does_not_return_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.anonymous_user)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user1_template.hash], request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_list_as_anonymous_does_not_return_global(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.global_template.hash], request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_list_as_user_returns_user_template(self):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user1_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_list_as_user_returns_global_template(self):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.global_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_list_as_user_does_not_return_other_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user2_template.hash], request=mock_request
        )
        self.assertTrue(templates.count() == 0)

    def test_get_all_accessible_by_hash_list_as_staff_returns_user_template(self):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user1_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_list_as_staff_returns_global_template(self):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.global_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_list_as_staff_does_not_return_other_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.staff_user1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user2_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_list_as_superuser_returns_user_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user1_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))

    def test_get_all_accessible_by_hash_list_as_superuser_returns_global_template(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.global_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template not in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))

    def test_get_all_accessible_by_hash_list_as_superuser_returns_other_user_template(
        self,
    ):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all_accessible_by_hash_list(
            [self.fixture.user2_template.hash], request=mock_request
        )
        self.assertTrue(self.fixture.user1_template not in list(templates))
        self.assertTrue(self.fixture.user2_template in list(templates))
        self.assertTrue(self.fixture.global_template not in list(templates))


class TestTemplateGetAll(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()
        self.template_id_list = [
            str(self.fixture.user1_template.id),
            str(self.fixture.user2_template.id),
            str(self.fixture.global_template.id),
        ]

    def test_get_all_as_anonymous_raises_acces_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.get_all(request=mock_request)

    def test_get_all_as_user_raises_acces_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.get_all(request=mock_request)

    def test_get_all_as_staff_raises_acces_control_error(self):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            template_api.get_all(request=mock_request)

    def test_get_all_as_superuser_returns_all_templates(self):
        mock_request = create_mock_request(user=self.superuser1)
        templates = template_api.get_all(request=mock_request)
        self.assertTrue(self.fixture.user1_template in list(templates))
        self.assertTrue(self.fixture.user2_template in list(templates))
        self.assertTrue(self.fixture.global_template in list(templates))


class TestTemplateDelete(MongoIntegrationBaseTestCase):

    fixture = fixture_template

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_delete_user_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.delete(self.fixture.user1_template, request=mock_request)

    def test_delete_global_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            template_api.delete(self.fixture.global_template, request=mock_request)

    def test_delete_own_template_as_user_saves(self):
        mock_request = create_mock_request(user=self.user1)
        template_api.delete(self.fixture.user1_template, request=mock_request)

    def test_delete_other_users_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.delete(self.fixture.user2_template, request=mock_request)

    def test_delete_global_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            template_api.delete(self.fixture.global_template, request=mock_request)

    def test_delete_own_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.delete(self.fixture.user1_template, request=mock_request)

    def test_delete_other_users_template_as_staff_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            template_api.delete(self.fixture.user2_template, request=mock_request)

    def test_delete_global_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        template_api.delete(self.fixture.global_template, request=mock_request)

    def test_delete_own_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.delete(self.fixture.user1_template, request=mock_request)

    def test_delete_other_users_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.delete(self.fixture.user2_template, request=mock_request)

    def test_delete_global_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        template_api.delete(self.fixture.global_template, request=mock_request)



""" Common forms
"""
from django import forms

from core_main_app.commons.validators import BlankSpacesValidator


class RenameForm(forms.Form):
    """Rename form"""

    field = forms.CharField(
        label="",
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": "Type the new name"}
        ),
        validators=[BlankSpacesValidator()],
    )

    id = forms.CharField(widget=forms.HiddenInput(), required=False)



"""Mailing task
"""
from __future__ import absolute_import

from logging import getLogger

from celery import shared_task
from django.core.mail import send_mail as django_send_mail, mail_admins, mail_managers
from django.template import loader

from core_main_app.settings import SERVER_EMAIL, EMAIL_SUBJECT_PREFIX

logger = getLogger(__name__)


@shared_task
def send_mail(
    recipient_list,
    subject,
    body,
    fail_silently=True,
    sender=SERVER_EMAIL,
):
    """Send email.

    Args:
        recipient_list:
        subject:
        body:
        fail_silently:
        sender:

    Returns:

    """
    try:

        # Send mail
        django_send_mail(
            subject=EMAIL_SUBJECT_PREFIX + subject,
            message="",
            from_email=sender,
            recipient_list=recipient_list,
            html_message=body,
            fail_silently=fail_silently,
        )
    except Exception as e:
        raise e


@shared_task
def send_mail_to_administrators(
    subject, path_to_template, context={}, fail_silently=True
):
    """Send email to administrators.

    Args:
        subject:
        path_to_template:
        context:
        fail_silently:

    Returns:

    """
    try:
        # Render the given template with context information
        template = loader.get_template(path_to_template)
        message = template.render(context)
        # Send mail
        mail_admins(
            subject=subject,
            message="",
            html_message=message,
            fail_silently=fail_silently,
        )
    except Exception as e:
        logger.warning(
            "send_mail_to_administrators threw an exception: ".format(str(e))
        )


@shared_task
def send_mail_to_managers(subject, path_to_template, context={}, fail_silently=True):
    """Send email to managers.

    Args:
        subject:
        path_to_template:
        context:
        fail_silently:

    Returns:

    """
    try:
        # Render the given template with context information
        template = loader.get_template(path_to_template)
        message = template.render(context)
        # Send mail
        mail_managers(
            subject=subject,
            message="",
            html_message=message,
            fail_silently=fail_silently,
        )
    except Exception as e:
        logger.warning("send_mail_to_managers throws an exception: ".format(str(e)))



"""
 Lock model
"""
import datetime
import threading

from bson.dbref import DBRef
from django_mongoengine import Document, fields
from mongoengine import errors as mongoengine_errors

from core_main_app.commons import exceptions

sem = threading.Semaphore()


class Lock(object):
    """
    Class Lock. Singleton thread safe.
    Only this object should be called to be used for an action regarding locking a Document.
    """

    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def acquire(cls):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = cls()
        sem.acquire()
        return cls.__singleton_instance

    @classmethod
    def release(cls):
        sem.release()

    def set_lock(self, object, user):
        """Set lock on a Document.

        Args:
            object:
            user:

        Returns:
        """
        database_lock_object = DatabaseLockObject()
        database_lock_object.object = DBRef(object._class_name, object.id)
        database_lock_object.user_id = str(user.id)
        database_lock_object.lock_date = datetime.datetime.now()
        database_lock_object.save()

    def remove_lock(self, database_lock_object):
        """Remove a lock.

        Args:
            database_lock_object:

        Returns:
        """
        database_lock_object.delete()

    def get_object_locked(self, object):
        """Get the locked object.

        Args:
            object:
        Returns:
        """
        try:
            return DatabaseLockObject.get_lock_by_object(object)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))


class DatabaseLockObject(Document):
    """
    Class DatabaseLockObject.
    """

    object = fields.ReferenceField(Document, blank=False)
    user_id = fields.StringField(blank=False)
    lock_date = fields.DateTimeField(blank=False)

    @staticmethod
    def get_lock_by_object(object):
        """Get lock relative to the given object.

        Args:
            object:

        Returns:

        """
        return DatabaseLockObject.objects.get(object=object)



"""Pagination utils for rest_framework package
"""
from urllib.parse import urlparse

from core_main_app.commons.exceptions import PaginationError
from core_main_app.utils.pagination.rest_framework_paginator.pagination import (
    StandardResultsSetPagination,
)


def get_page_number(url):
    """Get page number from url

    Args:
        url:

    Returns:

    """
    if url is None:
        return None

    try:
        # parse string url
        object_url = urlparse(url)
        # get query from url
        query_url = object_url.query
        # rest framework returns base (without 'page=') url when page is 1
        if "page=" not in query_url:
            return 1
        else:
            return int(query_url.split("=")[1])
    except Exception as e:
        return PaginationError(
            "An error occurred when getting page number from url: {}.".format(str(e))
        )


def get_paginator():
    """Create a paginator

    Returns:

    """
    # return paginator
    return StandardResultsSetPagination()



""" Access control fixture for views
"""
from core_main_app.components.blob.models import Blob
from core_main_app.components.data.models import Data
from core_main_app.components.template.models import Template
from core_main_app.components.workspace.models import Workspace
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class AccessControlDataFixture(FixtureInterface):
    """Access Control Data fixture"""

    USER_1_NO_WORKSPACE = 0
    USER_2_NO_WORKSPACE = 1
    USER_1_WORKSPACE_1 = 2
    USER_2_WORKSPACE_2 = 3

    template = None
    workspace_1 = None
    workspace_2 = None
    public_workspace = None
    data_collection = None
    data_1 = None
    data_2 = None
    data_workspace_1 = None
    data_public_workspace = None
    blob_1 = None
    blob_2 = None
    blob_workspace_1 = None
    blob_public_workspace = None
    blob_collection = None

    def insert_data(self):
        """Insert a set of Data.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template()
        self.generate_workspaces()
        self.generate_blob_collection()
        self.generate_data_collection()

    def generate_data_collection(self):
        """Generate a Data collection.

        Returns:

        """

        content = {"root": {"element": "value2"}}

        self.data_1 = Data(template=self.template, title="Data 1", user_id="1").save()
        self.data_2 = Data(template=self.template, title="Data 2", user_id="2").save()
        self.data_public_workspace = Data(
            template=self.template,
            title="Data Public Workspace",
            user_id="1",
            workspace=self.public_workspace.id,
            dict_content=content,
        ).save()
        self.data_workspace_1 = Data(
            template=self.template,
            title="Data Workspace 1",
            user_id="1",
            workspace=self.workspace_1.id,
            dict_content=content,
        ).save()
        self.data_collection = [
            self.data_1,
            self.data_2,
            self.data_public_workspace,
            self.data_workspace_1,
        ]

    def generate_blob_collection(self):
        """Generate a Data collection.

        Returns:

        """
        self.blob_1 = Blob(
            filename="blob1.txt",
            handle="http://127.0.0.1:8000/rest/blob/download/id1/",
            user_id="1",
        ).save()
        self.blob_2 = Blob(
            filename="blob2.txt",
            handle="http://127.0.0.1:8000/rest/blob/download/id2/",
            user_id="2",
        ).save()
        self.blob_public_workspace = Blob(
            filename="blob3.txt",
            handle="http://127.0.0.1:8000/rest/blob/download/id3/",
            user_id="1",
            workspace=self.public_workspace.id,
        ).save()
        self.blob_workspace_1 = Blob(
            filename="blob4.txt",
            handle="http://127.0.0.1:8000/rest/blob/download/id4/",
            user_id="1",
            workspace=self.workspace_1.id,
        ).save()
        self.blob_collection = [
            self.blob_1,
            self.blob_2,
            self.public_workspace,
            self.workspace_1,
        ]

    def generate_template(self):
        """Generate an unique Template.

        Returns:

        """
        template = Template()
        xsd = (
            '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
            '<xs:element name="tag"></xs:element></xs:schema>'
        )
        template.content = xsd
        template.hash = ""
        template.filename = "filename"
        self.template = template.save()

    def generate_workspaces(self):
        """Generate the workspaces.

        Returns:

        """
        self.workspace_1 = Workspace(
            title="Workspace 1", owner="1", read_perm_id="1", write_perm_id="1"
        ).save()
        self.workspace_2 = Workspace(
            title="Workspace 2", owner="2", read_perm_id="2", write_perm_id="2"
        ).save()
        self.public_workspace = Workspace(
            title="Public",
            owner="1",
            read_perm_id="3",
            write_perm_id="3",
            is_public=True,
        ).save()



"""XSL Transformation tag
"""

from django import template
from django.contrib.staticfiles import finders

from core_main_app.commons import exceptions
from core_main_app.components.template_xsl_rendering import (
    api as template_xsl_rendering_api,
)
from core_main_app.components.xsl_transformation import api as xsl_transformation_api
from core_main_app.settings import DEFAULT_DATA_RENDERING_XSLT
from core_main_app.utils.file import read_file_content
from core_main_app.utils.xml import xsl_transform

register = template.Library()


class XSLType(object):
    type_list = "List"
    type_detail = "Detail"


@register.simple_tag(name="xsl_transform_list")
def render_xml_as_html(*args, **kwargs):
    """Render an XML to HTML using the list xslt.
    Args:
        *args:
        **kwargs:

    Returns:
        HTML

    """
    xml_content = kwargs["xml_content"]
    template_id = kwargs.get("template_id", None)
    template_hash = kwargs.get("template_hash", None)
    xsl_transform_id = kwargs.get("xslt_id", None)
    request = kwargs.get("request", None)
    return _render_xml_as_html(
        xml_content,
        template_id,
        template_hash,
        XSLType.type_list,
        xsl_transform_id,
        request=request,
    )


@register.simple_tag(name="xsl_transform_detail")
def render_xml_as_html(*args, **kwargs):
    """Render an XML to HTML using the detail xslt.
    Args:
        *args:
        **kwargs:

    Returns:
        HTML

    """
    xml_content = kwargs["xml_content"]
    template_id = kwargs.get("template_id", None)
    template_hash = kwargs.get("template_hash", None)
    xsl_transform_id = kwargs.get("xslt_id", None)
    request = kwargs.get("request", None)
    return _render_xml_as_html(
        xml_content,
        template_id,
        template_hash,
        XSLType.type_detail,
        xsl_transform_id,
        request=request,
    )


def _render_xml_as_html(
    xml_string,
    template_id=None,
    template_hash=None,
    xslt_type=XSLType.type_list,
    xsl_transform_id=None,
    request=None,
):
    """Render an XML to HTML according to an xslt type (list or detail).
    Args:
        xml_string:
        template_id:
        template_hash:

        xslt_type:

    Returns:
        HTML

    """
    try:
        try:
            if xslt_type not in (XSLType.type_list, XSLType.type_detail):
                raise Exception("XSLT Type unknown. Default xslt will be used.")
            if xsl_transform_id:
                xsl_transformation = xsl_transformation_api.get_by_id(xsl_transform_id)
            elif template_id or template_hash:
                if template_id:
                    template_xsl_rendering = (
                        template_xsl_rendering_api.get_by_template_id(template_id)
                    )
                else:
                    template_xsl_rendering = (
                        template_xsl_rendering_api.get_by_template_hash(
                            template_hash, request=request
                        )
                    )

                if xslt_type == XSLType.type_list:
                    xsl_transformation = template_xsl_rendering.list_xslt
                else:
                    xsl_transformation = template_xsl_rendering.default_detail_xslt
            else:
                raise Exception(
                    "No template information provided. Default xslt will be used."
                )

            xslt_string = xsl_transformation.content

        except (Exception, exceptions.DoesNotExist):
            default_xslt_path = finders.find(DEFAULT_DATA_RENDERING_XSLT)
            xslt_string = read_file_content(default_xslt_path)

        return xsl_transform(xml_string, xslt_string)
    except Exception:
        return xml_string



"""Mailing util
"""
from django.template import loader
import core_main_app.utils.notifications.tasks.task_mail as task
from core_main_app.settings import SERVER_EMAIL, USE_BACKGROUND_TASK
from core_main_app.templatetags.stripjs import stripjs


def send_mail_from_template(
    recipient_list,
    subject,
    path_to_template,
    context={},
    fail_silently=True,
    sender=SERVER_EMAIL,
):
    """Send email.

    Args:
        recipient_list:
        subject:
        path_to_template:
        context:
        fail_silently:
        sender:

    Returns:

    """
    # Render the given template with context information
    template = loader.get_template(path_to_template)
    body = template.render(context)
    _send_email(
        recipient_list=recipient_list,
        subject=subject,
        body=body,
        fail_silently=fail_silently,
        sender=sender,
    )


def send_mail(
    recipient_list,
    subject,
    body,
    fail_silently=True,
    sender=SERVER_EMAIL,
):
    """Send email.

    Args:
        recipient_list:
        subject:
        body:
        fail_silently:
        sender:

    Returns:

    """

    stripped_body = stripjs(body)
    _send_email(
        recipient_list=recipient_list,
        subject=subject,
        body=stripped_body,
        fail_silently=fail_silently,
        sender=sender,
    )


def _send_email(recipient_list, subject, body, fail_silently, sender):
    """Send email async or sync from the rendered template

    Args:
        recipient_list:
        subject:
        body:
        fail_silently:
        sender:

    Returns:

    """
    if USE_BACKGROUND_TASK:
        # Async call. Use celery
        task.send_mail.apply_async(
            (
                recipient_list,
                subject,
                body,
                fail_silently,
                sender,
            ),
            countdown=1,
        )
    else:
        # Sync call
        task.send_mail(
            recipient_list=recipient_list,
            subject=subject,
            body=body,
            fail_silently=fail_silently,
            sender=sender,
        )


def send_mail_to_administrators(
    subject, path_to_template, context={}, fail_silently=True
):
    """Send email to administrators.

    Args:
        subject:
        path_to_template:
        context:
        fail_silently:

    Returns:

    """
    if USE_BACKGROUND_TASK:
        # Async call. Use celery
        task.send_mail_to_administrators.apply_async(
            (subject, path_to_template, context, fail_silently), countdown=1
        )
    else:
        # Sync call
        task.send_mail_to_administrators(
            subject, path_to_template, context, fail_silently
        )


def send_mail_to_managers(subject, path_to_template, context={}, fail_silently=True):
    """Send email to managers.

    Args:
        subject:
        path_to_template:
        context:
        fail_silently:

    Returns:

    """
    if USE_BACKGROUND_TASK:
        # Async call. Use celery
        task.send_mail_to_managers.apply_async(
            (subject, path_to_template, context, fail_silently), countdown=1
        )
    else:
        # Sync call
        task.send_mail_to_managers(subject, path_to_template, context, fail_silently)



""" App model to manage permissions.
"""
from django.db import models

from core_main_app.permissions import rights
from core_main_app.permissions.utils import get_formatted_name


class Main(models.Model):
    """Main model containing several information:

    Attributes:
        Meta.verbose_name (:py:class:`str`): Name of the app.
        Meta.permissions (:py:class:`list`): Default set of permissions bundled with the app.
    """

    class Meta(object):
        verbose_name = "core_main_app"
        default_permissions = ()
        permissions = (
            (rights.publish_data, get_formatted_name(rights.publish_data)),
            (rights.publish_blob, get_formatted_name(rights.publish_blob)),
        )



""" Unit tests of get_attribute templatetag
"""
from unittest.case import TestCase

from core_main_app.templatetags.get_attribute import get_attribute


class TestObject(object):
    test = "test"


class TestSystemGetAllExcept(TestCase):
    def test_get_attribute_on_empty_object(self):
        result = get_attribute({}, "test")
        self.assertEqual(result, "")

    def test_get_attribute_on_none(self):
        result = get_attribute(None, "test")
        self.assertEqual(result, "")

    def test_get_attribute_on_object(self):
        test_obj = TestObject()
        result = get_attribute(test_obj, "test")
        self.assertEqual(result, "test")

    def test_get_attribute_on_dict(self):
        test_dict = {"test": "test"}
        result = get_attribute(test_dict, "test")
        self.assertEqual(result, "test")

    def test_get_attribute_on_table(self):
        test_tab = ["test"]
        result = get_attribute(test_tab, "0")
        self.assertEqual(result, "test")



"""Template tag to check user permission
"""

from django import template

import core_main_app.permissions.rights as RIGHTS
from core_main_app.components.group import api as group_api

register = template.Library()


@register.filter(name="has_perm")
def has_perm(user, permission):
    """Check if user has the right level of permission to access a feature.

    Args:
        user:
        permission:

    Returns:

    """
    try:
        permission_split = permission.split(".")
        permission_name = permission_split[1]

        if user.is_anonymous:
            # We can give directly the permission name
            access = group_api.get_by_name_and_permission(
                name=RIGHTS.anonymous_group, permission_codename=permission_name
            )
        else:
            # We need to prefix with the app name
            access = user.has_perm(permission)
    except:
        # If something went wrong, we ask for an empty permission to give the access if it's a superUser
        access = user.has_perm("")

    return access



""" Fixture interface
"""


class FixtureInterface(object):
    """Represent the fixture interface with methods which must be implemented."""

    def insert_data(self):
        """Insert fixture's data

        Returns:

        """
        raise NotImplementedError()



"""Serializers used throughout the workspace Rest API
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_main_app.components.workspace import api as workspace_api
from core_main_app.components.workspace.models import Workspace


class WorkspaceSerializer(DocumentSerializer):
    """Workspace serializer"""

    class Meta(object):
        """Meta"""

        model = Workspace
        fields = ["id", "title", "owner", "is_public"]
        read_only_fields = ("id", "owner")

    def create(self, validated_data):
        """
        Create and return a new `workspace` instance, given the validated data.
        """
        return workspace_api.create_and_save(
            title=validated_data["title"],
            owner_id=validated_data["user"].id,
            is_public=validated_data["is_public"]
            if "is_public" in validated_data
            else False,
        )



""" Fixtures file for Templates
"""
from core_main_app.components.template.models import Template
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class AccessControlTemplateFixture(FixtureInterface):
    """Access Control Data fixture"""

    user1_template = None
    user2_template = None
    global_template = None
    template_collection = None

    def insert_data(self):
        """Insert a set of Data.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template_collection()

    def generate_template_collection(self):
        """Generate a Template Collections.

        Returns:

        """
        xsd = (
            '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
            '<xs:element name="tag"></xs:element></xs:schema>'
        )
        self.user1_template = Template(
            content=xsd,
            hash="user1_template_hash",
            filename="user1_template.xsd",
            user="1",
        ).save()
        self.user2_template = Template(
            content=xsd,
            hash="user2_template_hash",
            filename="user2_template.xsd",
            user="2",
        ).save()
        self.global_template = Template(
            content=xsd,
            hash="global_template_hash",
            filename="global_template.xsd",
            user=None,
        ).save()
        self.template_collection = [
            self.user1_template,
            self.user2_template,
            self.global_template,
        ]



""" discover rules for core main app
"""
import logging

import core_main_app.permissions.rights as rights

logger = logging.getLogger(__name__)


def init_rules(apps):
    """Init of group and permissions for the application.
    If the anonymous group does not exist, creation of the group
    If the default group does not exist, creation of the group

    Returns:

    """
    logger.info("START init rules.")

    try:
        group = apps.get_model("auth", "Group")

        # Get or Create the Group anonymous
        group.objects.get_or_create(name=rights.anonymous_group)

        # Get or Create the default group
        default_group, created = group.objects.get_or_create(name=rights.default_group)

        # Get curate permissions
        permission = apps.get_model("auth", "Permission")
        publish_data_perm = permission.objects.get(codename=rights.publish_data)
        publish_blob_perm = permission.objects.get(codename=rights.publish_blob)

        # Add permissions to default group
        default_group.permissions.add(publish_data_perm)
        default_group.permissions.add(publish_blob_perm)
    except Exception as e:
        logger.error("Impossible to init the rules: %s" % str(e))

    logger.info("FINISH init rules.")


def create_public_workspace():
    """Create and save a public workspace for registry. It will also create permissions.

    Returns:
    """
    logger.info("START create public workspace.")

    # We need the app to be ready to access the Group model
    from core_main_app.components.workspace import api as workspace_api
    from core_main_app.commons import exceptions

    try:
        try:
            # Test if global public workspace exists
            workspace_api.get_global_workspace()
        except exceptions.DoesNotExist:
            # Create workspace public global
            workspace_api.create_and_save("Global Public Workspace", is_public=True)
            logger.info("Public workspace created.")

    except Exception as e:
        logger.error("Impossible to create global public workspace: %s" % str(e))

    logger.info("FINISH create public workspace.")



""" Access control utils for HTTP Requests objects
"""
from django.http import HttpRequest
from rest_framework.request import Request

from core_main_app.commons.exceptions import CoreError


def get_request_from_args(*args, **kwargs):
    # get request from parameters
    request = kwargs["request"]
    if (
        isinstance(request, HttpRequest)
        or isinstance(request, Request)
        or request is None
    ):
        return request

    raise CoreError("request parameter improperly set")



""" Blob utils test class
"""
from unittest import TestCase

from core_main_app.commons.exceptions import CoreError
from core_main_app.utils.file import get_base_64_content_from_response
from mock import patch


class TestFileUtilsGetBase64ContentFromResponse(TestCase):
    def test_get_base_64_content_from_response_return_base_64_encoded_string(self):
        # Arrange
        response = MockResponse()
        # Act
        result = get_base_64_content_from_response(response)
        # Assert
        self.assertEqual(True, isinstance(result, str))

    def test_get_base_64_content_from_response_raise_core_exception_if_encoding_fails(
        self,
    ):
        # Arrange
        response = MockResponse()
        response.content = "this_is_a_string and it will fail"
        # Assert
        with self.assertRaises(CoreError):
            # Act
            get_base_64_content_from_response(response)

    @patch("base64.b64decode")
    def test_get_base_64_content_from_response_raise_core_exception_if_ascii_decoding_fails(
        self, mock_b64decode
    ):
        # Arrange
        response = MockResponse()
        response.content = "this_is_a_string and it will fail"
        mock_b64decode.side_effect = Exception()
        # Assert
        with self.assertRaises(CoreError):
            # Act
            get_base_64_content_from_response(response)


class MockResponse(object):
    content = b"my_string"



""" REST views for the template XSL rendering API
"""

from django.http import Http404
from rest_framework import status
from rest_framework.decorators import api_view, permission_classes
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.commons import exceptions
from core_main_app.components.template_xsl_rendering import (
    api as template_xsl_rendering_api,
)
from core_main_app.components.xsl_transformation import api as xsl_transformation_api
from core_main_app.rest.template_xsl_rendering.serializers import (
    TemplateXslRenderingSerializer,
)
from core_main_app.access_control.exceptions import AccessControlError


class TemplateXslRenderingList(APIView):
    """List all template XSL renderings, or create a new one"""

    permission_classes = (IsAuthenticated,)

    def get(self, request):
        """Get all templates XSL renderings

        Args:
            request: HTTP request

        Returns:
            - code: 200
              content: List of XSL renderings
            - code: 401
              content: Unauthorized
            - code: 500
              content: Internal server error
        """
        try:
            # Retrieve all objects
            template_xsl_rendering_list = template_xsl_rendering_api.get_all()

            # Serialize object
            serializer = TemplateXslRenderingSerializer(
                template_xsl_rendering_list, many=True
            )

            # Return response
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def post(self, request):
        """Create a XSL template rendering

        Parameters:
            {
                "template": "template_object_id",
                "list_xslt": "xslt_object_id",
                "default_detail_xslt": "xslt_object_id"
                "list_detail_xslt": "xslt_object_id"
            }

        Args:
            request: HTTP request

        Returns:
            - code: 201
              content: XSL rendering created
            - code: 400
              content: Validation error / not unique / model error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            serializer = TemplateXslRenderingSerializer(data=request.data)

            # Validate data
            serializer.is_valid(True)

            # Save data
            serializer.save()

            return Response(serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.ModelError as validation_exception:
            content = {"message": str(validation_exception)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.NotUniqueError as validation_exception:
            content = {"message": str(validation_exception)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except KeyError as validation_exception:
            content = {"message": validation_exception}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateXslRenderingDetail(APIView):
    """TemplateXslRendering details view"""

    permission_classes = (IsAuthenticated,)

    def get(self, request, pk):
        """Get `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get object
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)

            # Serialize object
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                template_xsl_rendering_object
            )

            # Return response
            return Response(template_xsl_rendering_serializer.data)
        except exceptions.DoesNotExist:
            content = {"message": "XSL rendering object not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def patch(self, request, pk):
        """Edit `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get object
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)

            # Build serializer
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                instance=template_xsl_rendering_object, data=request.data
            )

            # Validate data
            template_xsl_rendering_serializer.is_valid(True)
            # Save data
            template_xsl_rendering_serializer.save()

            return Response(
                template_xsl_rendering_serializer.data, status=status.HTTP_200_OK
            )
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {"message": "Data not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def delete(self, request, pk):
        """Delete a TemplateXSLRendering

        Args:
            request: HTTP request
            pk: ObjectId

        Returns:
            - code: 204
              content: Deletion successful
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)

            # delete object
            template_xsl_rendering_api.delete(template_xsl_rendering_object)

            # Return response
            return Response(status=status.HTTP_204_NO_CONTENT)
        except exceptions.DoesNotExist:
            content = {"message": "Template XSL rendering not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as ace:
            content = {"message": str(ace)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateXslRenderingAddDetailXslt(APIView):
    """TemplateXslRendering details view"""

    permission_classes = (IsAuthenticated,)

    def patch(self, request, pk, xslt_id):
        """Add detail xslt to `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId
            xslt_id : XsltObjectId

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get objects
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)
            xsl_transformation_object = xsl_transformation_api.get_by_id(xslt_id)

            # Add an xsl transformation
            template_xsl_rendering_api.add_detail_xslt(
                template_xsl_rendering_object, xsl_transformation_object
            )

            # Build serializer
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                instance=template_xsl_rendering_object
            )

            return Response(
                template_xsl_rendering_serializer.data, status=status.HTTP_200_OK
            )

        except Http404:
            content = {"message": "Data not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateXslRenderingSetDefaultDetailXslt(APIView):
    """TemplateXslRendering details view"""

    permission_classes = (IsAuthenticated,)

    def patch(self, request, pk, xslt_id):
        """Set default detail xslt to `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId
            xslt_id : XsltObjectId

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get objects
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)
            xsl_transformation_object = xsl_transformation_api.get_by_id(xslt_id)

            # Set default xsl transformation
            template_xsl_rendering_api.set_default_detail_xslt(
                template_xsl_rendering_object, xsl_transformation_object
            )

            # Build serializer
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                instance=template_xsl_rendering_object
            )

            return Response(
                template_xsl_rendering_serializer.data, status=status.HTTP_200_OK
            )
        except Http404:
            content = {"message": "Data not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateXslRenderingRemoveDetailXslt(APIView):
    """TemplateXslRendering details view"""

    permission_classes = (IsAuthenticated,)

    def patch(self, request, pk, xslt_id):
        """Remove a detail xslt from `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId
            xslt_id : XsltObjectId

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get objects
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)
            xsl_transformation_object = xsl_transformation_api.get_by_id(xslt_id)

            # Remove xsl transformation
            template_xsl_rendering_api.delete_detail_xslt(
                template_xsl_rendering_object, xsl_transformation_object
            )

            # Build serializer
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                instance=template_xsl_rendering_object
            )

            return Response(
                template_xsl_rendering_serializer.data, status=status.HTTP_200_OK
            )
        except Http404:
            content = {"message": "Data not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateXslRenderingSetListDetailXslt(APIView):
    """TemplateXslRendering details view"""

    permission_classes = (IsAuthenticated,)

    def patch(self, request, pk):
        """Add detail xslt to `TemplateXSLRendering` object from db

        Args:
            request: HTTP request
            pk: ObjectId

        Parameters:

            {
                "ids":
                     [
                        "xslt_id_1",
                        "xslt_id_2"
                     ]
            }

        Returns:
            TemplateXSLRendering
        """
        try:
            # Get objects
            template_xsl_rendering_object = template_xsl_rendering_api.get_by_id(pk)
            xsl_transformation_objects = xsl_transformation_api.get_by_id_list(
                request.data["ids"]
            )

            # Add an xsl transformation
            template_xsl_rendering_api.set_list_detail_xslt(
                template_xsl_rendering_object, xsl_transformation_objects
            )

            # Build serializer
            template_xsl_rendering_serializer = TemplateXslRenderingSerializer(
                instance=template_xsl_rendering_object
            )

            return Response(
                template_xsl_rendering_serializer.data, status=status.HTTP_200_OK
            )
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {"message": "Data not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)



""" Serializers used throughout the data Rest API for the admin user
"""

from core_main_app.components.data.models import Data
from core_main_app.rest.data.serializers import DataSerializer, XMLContentField
from core_main_app.components.data import api as data_api


class AdminDataSerializer(DataSerializer):
    """Admin Data serializer"""

    xml_content = XMLContentField()

    class Meta(object):
        """Meta"""

        model = Data
        fields = [
            "id",
            "template",
            "workspace",
            "user_id",
            "title",
            "xml_content",
            "creation_date",
            "last_modification_date",
            "last_change_date",
        ]
        read_only_fields = ("id",)

    def create(self, validated_data):
        """
        Create and return a new `Data` instance, given the validated data.
        """
        # Create data
        instance = Data(
            template=validated_data["template"],
            workspace=validated_data["workspace"]
            if "workspace" in validated_data
            else None,
            title=validated_data["title"],
            user_id=validated_data["user_id"]
            if "user_id" in validated_data
            else str(self.context["request"].user.id),
        )
        # Set XML content
        instance.xml_content = validated_data["xml_content"]
        # Set times
        instance.creation_date = validated_data.get("creation_date", None)
        instance.last_modification_date = validated_data.get(
            "last_modification_date", None
        )
        instance.last_change_date = validated_data.get("last_change_date", None)
        # Save the data
        data_api.admin_insert(instance, request=self.context["request"])
        # Encode the response body
        # NOTE: using xml_content property would update the last_modification_date
        instance._xml_content = validated_data["xml_content"].encode("utf-8")

        return instance



""" Version Manager Access Control
"""

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.version_manager.models import VersionManager
from core_main_app.settings import CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT
from core_main_app.utils.requests_utils.access_control import (
    get_request_from_args,
)


def can_read(func, document_id, request):
    """Can read document.

    Args:
        func:
        document_id:
        request:

    Returns:

    """
    request = get_request_from_args(document_id, request=request)

    # super user
    if request.user.is_superuser:
        return func(document_id, request=request)

    # get the document
    document = func(document_id, request=request)

    # anonymous user
    if request.user.is_anonymous:
        if document.user:
            raise AccessControlError(
                "Version Manager: The user doesn't have enough rights."
            )
        else:
            if not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT:
                raise AccessControlError(
                    "Version Manager: The user doesn't have enough rights."
                )

    # user is set
    if document.user and document.user != str(request.user.id):
        raise AccessControlError(
            "Version Manager: The user doesn't have enough rights."
        )

    return document


def can_write(func, *args, **kwargs):
    """Can write version_manager.

    Args:
        func:
        args:
        kwargs:

    Returns:

    """
    request = get_request_from_args(*args, **kwargs)

    # super user
    if request.user.is_superuser:
        return func(*args, **kwargs)

    # anonymous
    if request.user.is_anonymous:
        raise AccessControlError(
            "Version Manager: The user doesn't have enough rights."
        )

    version_manager = next(
        (arg for arg in args if isinstance(arg, VersionManager)), None
    )

    # user is set
    if version_manager.user:
        if version_manager.user == str(request.user.id):
            return func(*args, **kwargs)
    # user is not set
    else:
        if request.user.is_staff:
            return func(*args, **kwargs)

    raise AccessControlError("Version Manager: The user doesn't have enough rights.")


def can_read_list(func, list_id, request):
    """Can read list of version managers.

    Args:
        func:
        list_id:
        request:

    Returns:

    """
    # super user
    if request.user.is_superuser:
        return func(list_id, request=request)

    document_list = func(list_id, request)
    for version_manager in document_list:
        # anonymous user
        if request.user.is_anonymous:
            if version_manager.user:
                raise AccessControlError(
                    "Version Manager: The user doesn't have enough rights."
                )
            else:
                if not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT:
                    raise AccessControlError(
                        "Version Manager: The user doesn't have enough rights."
                    )

        # user is set
        if version_manager.user and version_manager.user != str(request.user.id):
            raise AccessControlError(
                "Version Manager: The user doesn't have enough rights."
            )

    return document_list


def can_read_global(func, *args, **kwargs):
    """Can read global version managers.

    Args:
        func:
        *args:
        **kwargs:

    Returns:

    """
    request = get_request_from_args(*args, **kwargs)
    if request.user.is_anonymous:
        if not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT:
            raise AccessControlError(
                "Version Manager: The user doesn't have enough rights."
            )
    return func(*args, **kwargs)


def can_read_version_manager(func, version_manager, version, request):
    """Can read global version managers.

    Args:
        func:
        version_manager:
        version:
        request:

    Returns:

    """
    request = get_request_from_args(version_manager, version, request=request)

    if request.user.is_superuser:
        return func(version_manager, version, request=request)
    # anonymous user
    if request.user.is_anonymous:
        if version_manager.user:
            raise AccessControlError(
                "Version Manager: The user doesn't have enough rights."
            )
        else:
            if not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT:
                raise AccessControlError(
                    "Version Manager: The user doesn't have enough rights."
                )

    # user is set
    if version_manager.user and version_manager.user != str(request.user.id):
        raise AccessControlError(
            "Version Manager: The user doesn't have enough rights."
        )

    return func(version_manager, version, request=request)



""" Integration Test for User Fixtures
"""

from core_main_app.components.user import api as user_api
from core_main_app.utils.integration_tests.integration_base_transaction_test_case import (
    MongoIntegrationTransactionTestCase,
)
from tests.components.user.fixtures.fixtures import UserFixtures


class TestUserFixtures(MongoIntegrationTransactionTestCase):
    """Test User fixtures"""

    def test_create_user(self):
        # Context
        user_count = len(user_api.get_all_users())
        self.assertEqual(user_count, 0)

        # Act
        UserFixtures().create_user(username="test username")

        # Assert
        list_user = user_api.get_all_users()
        self.assertEqual(list_user[0].username, "test username")
        self.assertEqual(len(list_user), user_count + 1)
        self.assertFalse(list_user[0].is_superuser)

    def test_create_super_user(self):
        # Context
        user_count = len(user_api.get_all_users())
        self.assertEqual(user_count, 0)

        # Act
        UserFixtures().create_super_user(username="test username")

        # Assert
        list_user = user_api.get_all_users()
        self.assertEqual(list_user[0].username, "test username")
        self.assertEqual(len(list_user), user_count + 1)
        self.assertTrue(list_user[0].is_superuser)



from django import forms

from core_main_app.commons.validators import BlankSpacesValidator


class UploadForm(forms.Form):
    """
    Form to upload.
    """

    name = forms.CharField(
        max_length=100,
        required=True,
        widget=forms.TextInput(attrs={"class": "form-control"}),
        validators=[BlankSpacesValidator()],
    )
    upload_file = forms.FileField(label="Select a file", required=True)



"""
    The Database pymongo tool contains the available function relative to database operation (connection)
"""
import logging
import re

from pymongo import MongoClient
from pymongo.errors import OperationFailure

from core_main_app.commons import exceptions

logger = logging.getLogger(__name__)


class Database(object):
    """Represent Database."""

    def __init__(self):
        self.client = None

    def connect(self, db_uri, db_name, doc_class=dict):
        """Connect to the database from settings.py.

        Args:
            db_uri:
            db_name:
            doc_class:

        Returns:

        """
        # create a connection
        self.client = MongoClient(db_uri, document_class=doc_class)
        db = self.client[db_name]  # connect to the database
        if db is not None:
            return db  # return db connection
        else:  # or raise an exception
            raise exceptions.CoreError("Database connection error")

    def close_connection(self):
        """
        Close the client connection.
        """
        self.client.close()

    @staticmethod
    def get_collection(db, collection_name):
        """Return cursor of collection name in parameters.

        Args:
            db:
            collection_name:

        Returns:

        """
        try:
            data_list = db[collection_name]  # get the data collection
            return data_list  # return collection
        except:  # or raise an exception
            raise exceptions.CoreError("Collection in database does not exist")

    @staticmethod
    def clean_database(db):
        """Clean the database.

        Args:
            db:

        Returns:

        """
        # clear all collections
        for collection in db.list_collection_names():
            try:
                if collection != "system.indexes":
                    db.drop_collection(collection)
            except OperationFailure as e:
                logger.warning("clean_database threw an exception: ".format(str(e)))


def get_full_text_query(text):
    """Return a full text query.

    Args:
        text: List of keywords

    Returns: The corresponding query

    """
    full_text_query = {}
    word_list = re.sub("[^\w]", " ", text, flags=re.UNICODE).split()
    word_list = ['"' + x + '"' for x in word_list]
    word_list = " ".join(word_list)
    if len(word_list) > 0:
        full_text_query = {"$text": {"$search": word_list}}

    return full_text_query



""" View builder for view data page.
"""
import logging

from core_main_app import settings
from core_main_app.commons import exceptions
from core_main_app.components.template_xsl_rendering import (
    api as template_xsl_rendering_api,
)
from bson import ObjectId


logger = logging.getLogger(__name__)


def build_page(data_object, display_admin_version=False):
    """Generic page building data

    Args:
        data_object:
        display_admin_version:

    Returns:
    """
    page_info = {
        "error": None,
        "context": dict(),
        "assets": dict(),
        "modals": list(),
    }

    try:
        display_xslt_selector = True
        try:
            template_xsl_rendering = template_xsl_rendering_api.get_by_template_id(
                data_object.template.id
            )
            xsl_transformation_id = (
                template_xsl_rendering.default_detail_xslt.id
                if template_xsl_rendering.default_detail_xslt
                else None
            )
            if not template_xsl_rendering.list_detail_xslt or (
                template_xsl_rendering.default_detail_xslt is not None
                and len(template_xsl_rendering.list_detail_xslt) == 1
            ):
                display_xslt_selector = False

            if xsl_transformation_id is not None:
                xsl_transformation_id = ObjectId(xsl_transformation_id)
        except Exception as exception:
            logger.warning(
                "An exception occured when retrieving XSLT: %s" % str(exception)
            )
            display_xslt_selector = False
            template_xsl_rendering = None
            xsl_transformation_id = None

        page_info["context"] = {
            "data": data_object,
            "share_pid_button": False,
            "template_xsl_rendering": template_xsl_rendering,
            "xsl_transformation_id": xsl_transformation_id,
            "can_display_selector": display_xslt_selector,
        }

        page_info["assets"] = {
            "js": [
                {"path": "core_main_app/common/js/XMLTree.js", "is_raw": False},
                {"path": "core_main_app/user/js/data/detail.js", "is_raw": False},
                {
                    "path": "core_main_app/user/js/data/change_display.js",
                    "is_raw": False,
                },
            ],
            "css": ["core_main_app/common/css/XMLTree.css"],
        }

        if "core_file_preview_app" in settings.INSTALLED_APPS:
            page_info["assets"]["js"].extend(
                [
                    {
                        "path": "core_file_preview_app/user/js/file_preview.js",
                        "is_raw": False,
                    }
                ]
            )
            page_info["assets"]["css"].append(
                "core_file_preview_app/user/css/file_preview.css"
            )
            page_info["modals"].append(
                "core_file_preview_app/user/file_preview_modal.html"
            )

        if (
            "core_linked_records_app" in settings.INSTALLED_APPS
            and not display_admin_version
            and settings.AUTO_SET_PID
        ):
            page_info["context"]["share_pid_button"] = True
            page_info["assets"]["js"].extend(
                [
                    {
                        "path": "core_main_app/user/js/sharing_modal.js",
                        "is_raw": False,
                    },
                    {
                        "path": "core_linked_records_app/user/js/sharing/data_detail.js",
                        "is_raw": False,
                    },
                ]
            )
            page_info["modals"].append(
                "core_linked_records_app/user/sharing/data_detail/modal.html"
            )
    except exceptions.DoesNotExist:
        page_info["error"] = "Data not found"
    except exceptions.ModelError:
        page_info["error"] = "Model error"
    except Exception as e:
        page_info["error"] = str(e)
    finally:
        return page_info


def render_page(request, render_function, template, context):
    if context["error"] is None:
        return render_function(
            request,
            template,
            context=context["context"],
            assets=context["assets"],
            modals=context["modals"],
        )
    else:
        return render_function(
            request,
            "core_main_app/common/commons/error.html",
            context={"error": context["error"]},
        )



""" Data tasks
"""

from bson.objectid import ObjectId
from celery import shared_task
from celery.result import AsyncResult

from core_main_app.components.data import api as data_api
from core_main_app.components.user import api as user_api
from core_main_app.system import api as system_api


@shared_task
def async_migration_task(data_list, template_id, user_id, migrate):
    """Async task which perform a migration / validation of the data list for the given target template id

    Args:
        data_list:
        template_id:
        user_id:
        migrate: (boolean) Perform the migration

    Return:
        {"valid": ["id"...], "wrong": ["id"...]}
    """
    success = []
    errors = []
    current_progress = 0
    total_data = len(data_list)

    try:
        user = user_api.get_user_by_id(user_id)
        target_template = system_api.get_template_by_id(template_id)

        for data_id in data_list:
            data = data_api.get_by_id(data_id, user=user)
            # modify the data temporarily with the new targeted template
            data.template = target_template

            try:
                # save the new template for the data if the migration is True
                if migrate:
                    system_api.upsert_data(data)
                else:
                    # check if the data is valid
                    data_api.check_xml_file_is_valid(data)

                success.append(str(data.id))
            except Exception as e:
                errors.append(str(data.id))
            finally:
                # increase the current progress and update the task state
                current_progress += 1
                async_migration_task.update_state(
                    state="PROGRESS",
                    meta={"current": current_progress, "total": total_data},
                )
    except Exception as e:
        async_migration_task.update_state(
            state="ABORT", meta={"current": current_progress, "total": total_data}
        )
        raise Exception(f"Something went wrong: {str(e)}")

    return {"valid": success, "wrong": errors}


@shared_task
def async_template_migration_task(templates, target_template_id, user_id, migrate):
    """Async task which perform a migration / validation of all the data which belong to the given template id list

    Args:
        templates:
        target_template_id:
        user_id
        migrate: (boolean) Perform the migration

    Return:
        {"valid": <number>, "wrong": <number>}
    """
    # get the data list to check
    current_data_progress = 0
    current_template_progress = -1
    total_data = 0
    total_template = len(templates)
    success = []
    error = []
    try:
        if target_template_id and total_template > 0:
            # get the user
            user = user_api.get_user_by_id(user_id)
            # get the target template
            target_template = system_api.get_template_by_id(target_template_id)

            for template_id in templates:

                # increase the number of processed template
                current_template_progress += 1
                # rest de number of data
                current_data_progress = 0

                # get a QuerySet of all the data with the given template
                data_list = data_api.execute_query(
                    {"template": ObjectId(template_id)}, user=user
                )

                total_data = data_list.count()

                # extract the data id from the list
                data_list_id = data_list.values_list("id")

                for data_id in data_list_id:

                    # get the data
                    data = data_api.get_by_id(data_id, user)

                    # modify the data temporarily with the new targeted template
                    data.template = target_template
                    # check if the data is valid
                    try:
                        # save the new template for the data if the migration is True
                        if migrate:
                            system_api.upsert_data(data)
                        else:
                            data_api.check_xml_file_is_valid(data)

                        success.append(str(data.id))
                    except Exception as e:
                        error.append(str(data.id))
                    finally:
                        # increase the current progress and update the task state
                        current_data_progress += 1
                        async_template_migration_task.update_state(
                            state="PROGRESS",
                            meta={
                                "template_current": current_template_progress,
                                "template_total": total_template,
                                "data_current": current_data_progress,
                                "data_total": total_data,
                            },
                        )

            return {"valid": success, "wrong": error}

        else:
            async_template_migration_task.update_state(
                state="ABORT",
                meta={
                    "template_current": current_template_progress,
                    "template_total": total_template,
                    "data_current": current_data_progress,
                    "data_total": total_data,
                },
            )
            raise Exception(
                "Wrong template id."
                if not target_template_id
                else "Please provide template id."
            )
    except Exception as e:
        async_template_migration_task.update_state(
            state="ABORT",
            meta={
                "template_current": current_template_progress,
                "template_total": total_data,
                "data_current": current_data_progress,
                "data_total": total_data,
            },
        )
        raise Exception(f"Something went wrong: {str(e)}")


def get_task_progress(task_id):
    """Get task status for the given task id

    Args:
        task_id:

    Return:
        {
            'state': PENDING | PROGRESS | SUCCESS,
            'details': result (for SUCCESS) | null (for PENDING) | { PROGRESS info }
        }
    """
    result = AsyncResult(task_id)
    response_data = {
        "state": result.state,
        "details": result.info,
    }
    return response_data


def get_task_result(task_id):
    """Get task result for the given task id

    Args:
        task_id:

    Return: {
                "valid": ["data_id_1", "data_id_2" ...],
                "wrong": ["data_id_3", "data_id_4" ...]
            }
    """
    result = AsyncResult(task_id).result
    return result



from django import template

register = template.Library()


@register.filter(name="get_attribute")
def get_attribute(value, arg):
    """@see https://stackoverflow.com/questions/844746/performing-a-getattr-style-lookup-in-a-django-template"""

    if hasattr(value, str(arg)):
        return getattr(value, arg)
    elif isinstance(value, dict) and arg in value:
        return value[arg]
    elif arg.isdigit() and len(value) > int(arg):
        return value[int(arg)]
    else:
        return ""



"""
    Xml operation test class
"""
from collections import OrderedDict
from unittest import TestCase

import core_main_app.commons.exceptions as exceptions
from core_main_app.utils.xml import raw_xml_to_dict, remove_lists_from_xml_dict


class TestRawToDict(TestCase):
    def test_raw_to_dict_valid(self):
        # Arrange
        raw_xml = "<root><test>Hello</test></root>"
        expected_dict = OrderedDict([("root", OrderedDict([("test", "Hello")]))])

        # Act
        xml_dict = raw_xml_to_dict(raw_xml)

        # Assert
        self.assertEquals(expected_dict, xml_dict)

    def test_raw_to_dict_throws_exception_when_invalid_xml(self):
        # Arrange
        raw_xml = "<root><test>Hello</test?</root>"

        # Act # Assert
        with self.assertRaises(exceptions.XMLError):
            raw_xml_to_dict(raw_xml)


class TestRemoveListsFromXmlDict(TestCase):
    def test_remove_lists_from_xml_dict_empty_dict_does_nothing(self):
        # Arrange
        xml_dict = {}
        # Act
        remove_lists_from_xml_dict(xml_dict)
        # Assert
        self.assertTrue(xml_dict == {})

    def test_remove_lists_from_xml_dict_root_list_is_removed(self):
        # Arrange
        xml_dict = {"list": [{"value": 1}, {"value": 2}]}
        # Act
        remove_lists_from_xml_dict(xml_dict)
        # Assert
        self.assertTrue(xml_dict == {})

    def test_remove_lists_from_xml_dict_root_list_is_not_removed_if_smaller_than_max_size(
        self,
    ):
        # Arrange
        xml_dict = {"list": [{"value": 1}, {"value": 2}]}
        # Act
        remove_lists_from_xml_dict(xml_dict, 3)
        # Assert
        self.assertTrue(xml_dict == {"list": [{"value": 1}, {"value": 2}]})

    def test_remove_lists_from_xml_dict_sub_element_list_is_removed(self):
        # Arrange
        xml_dict = {"root": {"list": [{"value": 1}, {"value": 2}]}}
        # Act
        remove_lists_from_xml_dict(xml_dict)
        # Assert
        self.assertTrue(xml_dict == {"root": {}})

    def test_remove_lists_from_xml_dict_sub_element_list_is_not_removed_if_smaller_than_max_size(
        self,
    ):
        # Arrange
        xml_dict = {"root": {"list": [{"value": 1}, {"value": 2}]}}
        # Act
        remove_lists_from_xml_dict(xml_dict, 3)
        # Assert
        self.assertTrue(xml_dict == {"root": {"list": [{"value": 1}, {"value": 2}]}})

    def test_remove_lists_from_xml_dict_sub_elements_only_list_is_removed(self):
        # Arrange
        xml_dict = {
            "root": {
                "list": [{"value": 1}, {"value": 2}],
                "int": 3,
                "str": "test",
                "dict": {"value": "test"},
            }
        }
        # Act
        remove_lists_from_xml_dict(xml_dict)
        # Assert
        self.assertTrue(
            xml_dict == {"root": {"int": 3, "str": "test", "dict": {"value": "test"}}}
        )



""" Data model
"""

from django_mongoengine import fields
from mongoengine.queryset.visitor import Q
from mongoengine import errors as mongoengine_errors
from mongoengine.queryset.base import NULLIFY

from core_main_app.commons import exceptions
from core_main_app.components.abstract_data.models import AbstractData
from core_main_app.components.template.models import Template
from core_main_app.components.workspace.models import Workspace


# TODO: Create publication workflow manager
# TODO: execute_query / execute_query_full_result -> use find method (RETURN FULL OBJECT)


class Data(AbstractData):
    """Data object"""

    template = fields.ReferenceField(Template, blank=False)
    user_id = fields.StringField()
    workspace = fields.ReferenceField(
        Workspace, reverse_delete_rule=NULLIFY, blank=True
    )

    meta = {"indexes": ["title", "last_modification_date", "template", "user_id"]}

    @staticmethod
    def get_all(order_by_field):
        """Get all data.

        Args:
            order_by_field: Order by field.

        Returns:

        """
        return Data.objects.order_by(*order_by_field)

    @staticmethod
    def get_all_except(order_by_field, id_list=None):
        """Get all data except for the ones with ID within the provided list.

        Args:
            id_list:
            order_by_field:

        Returns:
        """
        if id_list is None:
            return Data.get_all(order_by_field)

        return Data.objects(pk__nin=id_list).order_by(*order_by_field)

    @staticmethod
    def get_all_by_user_id(user_id, order_by_field):
        """Get all data relative to the given user id

        Args:
            user_id:
            order_by_field: Order by field.

        Returns:

        """
        return Data.objects(user_id=str(user_id)).order_by(*order_by_field)

    @staticmethod
    def get_all_except_user_id(user_id, order_by_field):
        """Get all data non relative to the given user id

        Args:
            user_id:
            order_by_field:

        Returns:

        """
        return Data.objects(user_id__nin=str(user_id)).order_by(*order_by_field)

    @staticmethod
    def get_all_by_id_list(list_id, order_by_field):
        """Return the object with the given list id.

        Args:
            list_id:
            order_by_field:

        Returns:
            Object collection
        """
        return Data.objects(pk__in=list_id).order_by(*order_by_field)

    @staticmethod
    def get_by_id(data_id):
        """Return the object with the given id.

        Args:
            data_id:

        Returns:
            Data (obj): Data object with the given id

        """
        try:
            return Data.objects.get(pk=str(data_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def execute_query(query, order_by_field):
        """Execute a query.

        Args:
            query:
            order_by_field: Order by field.

        Returns:

        """
        return Data.objects(__raw__=query).order_by(*order_by_field)

    @staticmethod
    def get_all_by_workspace(workspace, order_by_field):
        """Get all data that belong to the workspace.

        Args:
            workspace:
            order_by_field:

        Returns:

        """
        return Data.objects(workspace=workspace).order_by(*order_by_field)

    @staticmethod
    def get_all_by_list_workspace(list_workspace, order_by_field):
        """Get all data that belong to the list of workspace.

        Args:
            list_workspace:
            order_by_field:

        Returns:

        """
        return Data.objects(workspace__in=list_workspace).order_by(*order_by_field)

    @staticmethod
    def get_all_by_list_template(list_template, order_by_field):
        """Get all data that belong to the list of template.

        Args:
            list_template:
            order_by_field:

        Returns:

        """
        return Data.objects(template__in=list_template).order_by(*order_by_field)

    @staticmethod
    def get_all_by_user_and_workspace(user_id, list_workspace, order_by_field):
        """Get all data that belong to the list of workspace and owned by a user.

        Args:
            list_workspace:
            user_id:
            order_by_field:

        Returns:

        """
        return Data.objects(
            Q(workspace__in=list_workspace) | Q(user_id=str(user_id))
        ).order_by(*order_by_field)

    @staticmethod
    def get_all_by_templates_and_workspaces(
        list_template, list_workspace, order_by_field
    ):
        """Get all data stored in the list of workspace and created from the
        list of templates.

        Args:
            list_template:
            list_workspace:
            order_by_field:

        Returns:

        """
        return Data.objects(
            Q(workspace__in=list_workspace) & Q(template__in=list_template)
        ).order_by(*order_by_field)

    @staticmethod
    def aggregate(pipeline):
        """Execute an aggregate on the Data collection.

        Args:
            pipeline:

        Returns:

        """
        return Data.objects.aggregate(*pipeline)

    @staticmethod
    def get_none():
        """Return None object, used by data.

        Returns:

        """
        return Data.objects().none()



""" Integration Test for User Rest API
"""
from rest_framework.status import HTTP_404_NOT_FOUND

from core_main_app.utils.integration_tests.integration_base_transaction_test_case import (
    MongoIntegrationTransactionTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from tests.components.user.fixtures.fixtures import UserFixtures
from core_main_app.rest.user import views as user_rest_views
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)

from core_main_app.utils.tests_tools.RequestMock import RequestMock

user_fixture = UserFixtures()


class TestUserDetail(MongoIntegrationTransactionTestCase):
    def setUp(self):
        super(TestUserDetail, self).setUp()

    def test_get_returns_correct_user(self):
        # Arrange
        user = user_fixture.create_user()
        mock_user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_get(
            user_rest_views.UserDetail.as_view(), mock_user, param={"pk": user.id}
        )

        # Assert
        self.assertEqual(user.id, response.data["id"])

    def test_get_returns_incorrect_user_raises_404(self):
        # Arrange
        user = user_fixture.create_user()
        mock_user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_get(
            user_rest_views.UserDetail.as_view(), mock_user, param={"pk": user.id}
        )

        # Assert
        self.assertNotEqual(response.status_code, HTTP_404_NOT_FOUND)



""" Workspace access control
"""
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.permissions import api as permissions_api, rights as rights
from core_main_app.settings import CAN_SET_PUBLIC_DATA_TO_PRIVATE


def is_workspace_owner_to_perform_action_for_others(func, workspace, new_user_id, user):
    """Check if the user is the owner of the workspace to perform action for other user.

    Args:
        func
        workspace
        new_user_id
        user

    Returns:

    """
    if user.is_superuser:
        return func(workspace, new_user_id, user)

    _check_is_owner_workspace(workspace, user)
    return func(workspace, new_user_id, user)


def can_user_set_workspace_public(func, workspace, user):
    """Check if the user is the owner of the workspace.

    Args:
        func:
        workspace:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(workspace, user)

    _check_is_owner_workspace(workspace, user)

    publish_perm = permissions_api.get_by_codename(rights.publish_data)
    if not user.has_perm(
        publish_perm.content_type.app_label + "." + publish_perm.codename
    ):
        raise AccessControlError(
            "You don't have enough rights to set public this workspace."
        )

    return func(workspace, user)


def is_workspace_owner(func, workspace, user):
    """Check if the user is the owner of the workspace.

    Args:
        func:
        workspace:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(workspace, user)

    _check_is_owner_workspace(workspace, user)
    return func(workspace, user)


def can_delete_workspace(func, workspace, user):
    """Can user delete a workspace.

    Args:
        func:
        workspace:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(workspace, user)

    _check_is_owner_workspace(workspace, user)

    if CAN_SET_PUBLIC_DATA_TO_PRIVATE is False:
        if workspace.is_public:
            raise AccessControlError("The workspace can not be deleted.")

    return func(workspace, user)


def _check_is_owner_workspace(workspace, user):
    """Check that user is the owner of the workspace.

    Args:
        workspace:
        user:

    Returns:

    """
    if workspace.owner != str(user.id):
        raise AccessControlError(
            "The user does not have the permission. The user is not the owner of this workspace."
        )
